#!/usr/bin/env python
import os
import sys
import datetime
import argparse
import pika
import tempfile
import subprocess
import logging
# import tesserocr
import cv2
import numpy as np
from distutils.version import LooseVersion


if sys.version_info > (3, 0):
    import configparser
else:
    import ConfigParser

    reload(sys)
    sys.setdefaultencoding('utf8')

parser = argparse.ArgumentParser()
parser.add_argument('-c', '--config', help='use custom config file (edokumenty.ini is default)')


queueName = 'OCR'
args = parser.parse_args()

logging.basicConfig(format='%(asctime)s.%(msecs)d [%(process)d] %(levelname)s: %(message)s', datefmt='%H:%M:%S', filename='logs/ocr/' + datetime.datetime.today().strftime('%Y-%m-%d') + '.log', level=logging.INFO)


if args.config is not None:
    configFile = args.config
else:
    configFile = "edokumenty.ini"

logging.debug("Loading config file " + configFile)

# Wczytanie konfiguracji
try:
    if sys.version_info > (3, 0):
        Config = configparser.ConfigParser({'log_level': 'INFO'})
    else:
        Config = ConfigParser.ConfigParser()

    Config.read(configFile)

    # konfiguracja polaczenia z RabbitMQ
    pikaHost = Config.get("RabitMQ", "host")
    pikaVHost = Config.get("RabitMQ", "vhost")
    pikaPort = Config.getint("RabitMQ", "port")
    pikaCredentials = pika.PlainCredentials(Config.get("RabitMQ", "user"), Config.get("RabitMQ", "password"))
except Exception as e:
    logging.error("Incorrect config file - check " + configFile + ": " + str(e))
    sys.exit()

if Config.has_option("RabitMQ", 'log_level'):
    logLevel = getattr(logging, Config.get("RabitMQ", 'log_level'), logging.getLevelName(logging.INFO))
else:
    logLevel = logging.INFO

logging.getLogger().setLevel(logLevel)

connection = pika.BlockingConnection(pika.ConnectionParameters(pikaHost, pikaPort, pikaVHost, pikaCredentials))
channel = connection.channel()
channel.queue_declare(queue=queueName, durable=True)


def file_get_contents(filename):
    with open(filename, 'rb') as f:
        return f.read()


def deskew(im, max_skew=10):
    height, width = im.shape[:2]

    # Create a grayscale image and denoise it
    # im_gs = cv2.cvtColor(im, cv2.COLOR_BGR2GRAY)

    # below func takes very long time
    # im_gs = cv2.fastNlMeansDenoising(im_gs, h=3)
    # cv2.imwrite('a1a1.png', im_gs)
    # im_bw = cv2.threshold(im_gs, 0, 255, cv2.THRESH_BINARY_INV | cv2.THRESH_OTSU)[1]

    # Create an inverted B&W copy using Otsu (automatic) thresholding
    im_bw = cv2.threshold(im, 0, 255, cv2.THRESH_BINARY_INV | cv2.THRESH_OTSU)[1]

    # Detect lines in this image. Parameters here mostly arrived at by trial and error.
    lines = cv2.HoughLinesP(
        im_bw, 1, np.pi / 180, 150, minLineLength=width / 10, maxLineGap=width / 100
        # im_bw, 1, np.pi / 180, 200, minLineLength = width / 12, maxLineGap = width / 150
    )

    # Collect the angles of these lines (in radians)
    angles = []
    for line in lines:
        x1, y1, x2, y2 = line[0]
        angles.append(np.arctan2(y2 - y1, x2 - x1))

    # If the majority of our lines are vertical, this is probably a landscape image
    landscape = np.sum([abs(angle) > np.pi / 4 for angle in angles]) > len(angles) / 2

    # Filter the angles to remove outliers based on max_skew
    if landscape:
        angles = [
            angle
            for angle in angles
            if np.deg2rad(90 - max_skew) < abs(angle) < np.deg2rad(90 + max_skew)
        ]
    else:
        angles = [angle for angle in angles if abs(angle) < np.deg2rad(max_skew)]

    if len(angles) < 5:
        # Insufficient data to deskew
        return im

    angleMedian = np.median(angles)

    # Average the angles to a degree offset
    angle_deg = np.rad2deg(angleMedian)

    # If this is landscape image, rotate the entire canvas appropriately
    if landscape:
        if angle_deg < 0:
            im = cv2.rotate(im, cv2.ROTATE_90_CLOCKWISE)
            angle_deg += 90
        elif angle_deg > 0:
            im = cv2.rotate(im, cv2.ROTATE_90_COUNTERCLOCKWISE)
            angle_deg -= 90

    height, width = im.shape[:2]
    image_center = (width / 2, height / 2)

    # Rotate the image by the residual offset
    M = cv2.getRotationMatrix2D(image_center, angle_deg, 1)

    # abs_cos = abs(M[0, 0])
    # abs_sin = abs(M[0, 1])
    #
    # # find the new width and height bounds
    # bound_w = int(height * abs_sin + width * abs_cos)
    # bound_h = int(height * abs_cos + width * abs_sin)
    #
    # # subtract old image center (bringing image back to origo) and adding the new image center coordinates
    # M[0, 2] += bound_w / 2 - image_center[0]
    # M[1, 2] += bound_h / 2 - image_center[1]

    im = cv2.warpAffine(im, M, (width, height), flags=cv2.INTER_CUBIC, borderMode=cv2.BORDER_REPLICATE)
    return im


def removeLines(im):
    thresh = cv2.threshold(im, 0, 255, cv2.THRESH_BINARY_INV + cv2.THRESH_OTSU)[1]

    # Remove horizontal lines
    horizontal_kernel = cv2.getStructuringElement(cv2.MORPH_RECT, (40, 1))
    remove_horizontal = cv2.morphologyEx(thresh, cv2.MORPH_OPEN, horizontal_kernel, iterations=2)
    cnts = cv2.findContours(remove_horizontal, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
    cnts = cnts[0] if len(cnts) == 2 else cnts[1]
    for c in cnts:
        cv2.drawContours(im, [c], -1, (255, 255, 255), 5)

    # Remove vertical lines
    vertical_kernel = cv2.getStructuringElement(cv2.MORPH_RECT, (1, 40))
    remove_vertical = cv2.morphologyEx(thresh, cv2.MORPH_OPEN, vertical_kernel, iterations=2)
    cnts = cv2.findContours(remove_vertical, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
    cnts = cnts[0] if len(cnts) == 2 else cnts[1]
    for c in cnts:
        cv2.drawContours(im, [c], -1, (255, 255, 255), 5)


def callback(ch, method, properties, body):
    try:
        ocrParams = properties.headers

        logging.info('START OCR task for {},{} ...'.format(ocrParams['entity'], ocrParams['fileid']))
        logging.debug("headers: {0!r}".format(ocrParams))

        uri = ''
        delFileAfter = False

        if properties.content_type == 'application/pdf':
            logging.debug("creating tmp file")
            with tempfile.NamedTemporaryFile(suffix='.pdf', delete=False) as tmp:
                tmp.write(body)
                tmp.close()
                delFileAfter = True
                uri = tmp.name

        elif 'uri' in ocrParams:
            uri = os.path.normpath(ocrParams.get('uri'))

        if uri == '':
            # ch.basic_ack(delivery_tag=method.delivery_tag)
            raise Exception('task removed from queue (no URI or content)')

        logging.debug("processing file: " + uri)

        try:
            pngFileBase = os.path.join(tempfile.tempdir, tempfile.gettempprefix() + str(ocrParams.get('fileid')))
            pngFile = pngFileBase + '.png'
            imgCmd = ['pdftoppm', uri, pngFileBase, '-png', '-gray', '-r', '300', '-singlefile']
            logging.debug("call: " + ' '.join(imgCmd))
            subprocess.call(imgCmd)

            if not os.path.exists(pngFile):
                raise Exception('task removed from queue (no URI or content)')

            try:
                image = cv2.imread(pngFile, cv2.IMREAD_GRAYSCALE)
                deskewed_image = deskew(image)
                del image

                if logging.getLogger().level == logging.DEBUG:
                    cv2.imwrite(pngFileBase + '_orig_.png', deskewed_image)
                    cv2.imwrite(pngFileBase + '_deskew_.png', deskewed_image)

                removeLines(deskewed_image)

                if logging.getLogger().level == logging.DEBUG:
                    cv2.imwrite(pngFileBase + '_lines_.png', deskewed_image)

                cv2.imwrite(pngFile, deskewed_image)
                del deskewed_image

            except Exception as ce:
                logging.error(str(ce))

            mode = 'tsv'
            resultFile = pngFile + '.' + mode
            ocrCmd = ['tesseract', pngFile, pngFile, '--oem', '1', '--psm', '6', '-c', 'tessedit_create_tsv=1', '-l', 'Latin']

            logging.debug("call: " + ' '.join(ocrCmd))

            subprocess.call(ocrCmd)
            if logging.getLogger().level != logging.DEBUG:
                os.remove(pngFile)

            if properties.reply_to != '':
                logging.info("replay_to: " + properties.reply_to)

                ch.basic_publish(exchange='', routing_key=properties.reply_to,
                                 properties=pika.BasicProperties(content_type='text/tsv', correlation_id=properties.correlation_id, headers=ocrParams),
                                 body=file_get_contents(resultFile))

            if delFileAfter and os.path.exists(uri):
                os.remove(uri)

            ch.basic_ack(delivery_tag=method.delivery_tag)

            logging.info("...DONE " + ocrParams.get('fileid'))

        except Exception as ee:
            ch.basic_reject(delivery_tag=method.delivery_tag, requeue=False)
            logging.error(str(ee))

    except Exception as e2:
        ch.basic_reject(delivery_tag=method.delivery_tag, requeue=False)
        logging.error(str(e2))


channel.basic_qos(prefetch_count=1)
if LooseVersion(pika.__version__) >= LooseVersion('1.0'):
    channel.basic_consume(on_message_callback=callback, queue=queueName)
else:
    channel.basic_consume(consumer_callback=callback, queue=queueName)


logging.info('Waiting for OCR tasks. To exit press CTRL+C')
try:
    channel.start_consuming()
except KeyboardInterrupt:
    logging.info("interrupt received, stopping...")
finally:
    # clean up
    channel.close()
    sys.exit(0)
