#!/usr/bin/env python
import os
import sys
import datetime
import argparse
import pika
import subprocess
import logging


if sys.version_info > (3, 0):
    import configparser
else:
    import ConfigParser

    reload(sys)
    sys.setdefaultencoding('utf8')

parser = argparse.ArgumentParser()
parser.add_argument('-c', '--config', help='use custom config file (edokumenty.ini is default)')


queueName = 'OCR_RESULT'
args = parser.parse_args()

logging.basicConfig(format='%(asctime)s.%(msecs)d [%(process)d] %(levelname)s: %(message)s', datefmt='%H:%M:%S', filename='logs/ocr_result/' + datetime.datetime.today().strftime('%Y-%m-%d') + '.log', level=logging.INFO)

if args.config is not None:
    configFile = args.config
else:
    configFile = "edokumenty.ini"

logging.debug("Loading config file " + configFile)

# Wczytanie konfiguracji
try:
    if sys.version_info > (3, 0):
        Config = configparser.ConfigParser({'log_level': 'INFO'})
    else:
        Config = ConfigParser.ConfigParser()

    Config.read(configFile)

    # konfiguracja polaczenia z RabbitMQ
    pikaHost = Config.get("RabitMQ", "host")
    pikaVHost = Config.get("RabitMQ", "vhost")
    pikaPort = Config.getint("RabitMQ", "port")
    pikaCredentials = pika.PlainCredentials(Config.get("RabitMQ", "user"), Config.get("RabitMQ", "password"))

    # edokLocation = Config.get("edokumenty", "location")
    # ocrPHPScript = Config.get("edokumenty", "ocr_result_script")

except Exception as e:
    logging.error("Incorrect config file - check " + configFile + ": " + str(e))
    sys.exit()

if Config.has_option("RabitMQ", 'log_level'):
    logLevel = getattr(logging, Config.get("RabitMQ", 'log_level'), logging.getLevelName(logging.INFO))
else:
    logLevel = logging.INFO

logging.getLogger().setLevel(logLevel)

connection = pika.BlockingConnection(pika.ConnectionParameters(pikaHost, pikaPort, pikaVHost, pikaCredentials))
channel = connection.channel()
channel.queue_declare(queue=queueName, durable=True)


def callback(ch, method, properties, body):
    try:
        ocrParams = properties.headers

        if properties.content_type != 'text/tsv':
            raise Exception('wrong content_type (expecting text/tsv)')

        if ('fileid' not in ocrParams) or ('entity' not in ocrParams) or ('ent_id' not in ocrParams) or ('ready_uri' not in ocrParams):
            raise Exception('missing required headers (fileid,entity,ent_id,ready_uri)')

        logging.info('START OCR_RESULT task for {}:{},{} ...'.format(ocrParams['entity'], ocrParams['ent_id'], ocrParams['fileid']))
        logging.debug('headers: {0!r}'.format(ocrParams))

        try:
            ocrResScript = ocrParams.get('ready_uri') + '/apps/edokumenty/modules/OCR/services/ocr_res_worker.php'
            ocrResScript = os.path.normpath(ocrResScript)
            if not os.path.exists(ocrResScript):
                raise Exception('directory does not exist (ready_uri)')

            phpCmd = ['php', ocrResScript, (ocrParams.get('entity') or '""'), ocrParams.get('ent_id'), ocrParams.get('fileid')]

            logging.debug("call: " + ' '.join(phpCmd))

            process = subprocess.Popen(phpCmd, cwd=ocrParams.get('ready_uri'), stdin=subprocess.PIPE, stdout=subprocess.PIPE)
            process.stdin.write(body)
            result = process.communicate()[0].decode("utf-8")
            process.stdin.close()

            if (result != '') and (result != ocrParams.get('fileid')):
                # todo: requeue
                ch.basic_reject(delivery_tag=method.delivery_tag, requeue=False)
                logging.error('Result mismatch: ' + result + ' != ' + ocrParams.get('fileid'))
            else:
                ch.basic_ack(delivery_tag=method.delivery_tag)
                logging.info("...DONE " + ocrParams.get('fileid'))

        except subprocess.CalledProcessError as pe:
            # todo: requeue
            ch.basic_reject(delivery_tag=method.delivery_tag, requeue=False)
            logging.error("output = %s, error code = %s\n" % (pe.output, pe.returncode))

        except Exception as ee:
            ch.basic_reject(delivery_tag=method.delivery_tag, requeue=False)
            logging.error(str(ee))

    except Exception as e2:
        ch.basic_reject(delivery_tag=method.delivery_tag, requeue=False)
        logging.error(str(e2))


channel.basic_qos(prefetch_count=1)
channel.basic_consume(on_message_callback=callback, queue=queueName)

logging.info('Waiting for OCR_RESULT tasks...')
try:
    channel.start_consuming()

except KeyboardInterrupt:
    logging.info("interrupt received, stopping...")
finally:
    # clean up
    channel.close()
    sys.exit(0)
